// recoderd: a simple server using mplayer 
//           to transcode a file to MPEG for
//           DVB playback
//
// (c) 2004 Patrick Cernko <errror@gmx.de>

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <unistd.h>    // pipe(), close(), read(), write(), exit(), execl(), select()
#include <fcntl.h>     // fcntl()
#include <sys/time.h>  // select()
#include <sys/types.h> // select()
#include <sys/wait.h>  // waitpid()
#include <errno.h>     // errno
#include <stdio.h>     // snprintf()

#include <iostream>
#include <string>
#include <memory>

// includes from the toolbox lib
#include "cfile.h"
#include "csocket.h"
#include "cselect.h"


#ifndef DEFAULT_CMD_PORT
#  warning DEFAULT_CMD_PORT unset, Using 40404 as port
#  define DEFAULT_CMD_PORT 40404
#endif

#ifndef DEFAULT_DATA_PORT
#  warning DEFAULT_DATA_PORT unset, Using DEFAULT_CMD_PORT+1 as port
#  define DEFAULT_DATA_PORT (DEFAULT_CMD_PORT+1)
#endif


using namespace std;

int main(int /*argc*/, char* /*argv*/[]) {
  int cmd_port = DEFAULT_CMD_PORT;
  int data_port = DEFAULT_DATA_PORT;

  // parse args


  // used objects
  cTBSocket net_listener;    // a socket listener
  cTBSelect io_select;       // a select() object

  int inpipe[2], outpipe[2]; // pipes for the std{in,out,err} connection to mplayer
  int pid(0);                // the pid of our mplayer child

  cTBFile mp_input;          // mplayer's input is a output for us
  cTBFile mp_output;         // mplayer's output is a input for us
  cTBFile fifo_input;        // mplayer's data output via the fifo
  cTBFile std_input;         // stdin as TB-object for TB's select


  // create a fifo for the data output from mplayer
  char tmp_dir_tmpl[25];
  strncpy(tmp_dir_tmpl, "/tmp/recoderd.XXXXXX", 25);
  char* tmp_dir(mkdtemp(tmp_dir_tmpl));
  if (tmp_dir == NULL) {
    cerr << "ERROR: Could not create temp. dir: " << strerror(errno) << endl;
    return 1;
  }
  if (chdir(tmp_dir) == -1) {
    cerr << "ERROR: Could not change to temp. dir \"" << tmp_dir << "\": " << strerror(errno) << endl;
    return 1;
  }
  if (mkfifo("grab.mpg", 0600) == -1) {
    cerr << "ERROR: Problems creating fifo grab.mpg: " << strerror(errno) << endl;
    unlink("grab.mpg");
    rmdir(tmp_dir);
    return 1;
  }


  // listen and accept a connection
  // 1. listen
  if (!net_listener.Listen(cmd_port, 0)) {
    cerr << "ERROR: Problems opening server port: " << strerror(errno) << endl;
    return 1;
  }
  // 2. wait for incomming connection
  do {
    io_select.Clear();
    io_select.Add(net_listener, false /* waiting for input*/);
    io_select.Select(100); // wait 100ms
  } while (!io_select.CanRead(net_listener));
  // 3. accept the connection
  auto_ptr<cTBSocket> net_cmd(net_listener.Accept());
  while (!net_cmd.get()) {
    cout << "Problems accepting incoming connection: " << strerror(errno) << endl;
    net_cmd.reset(net_listener.Accept());
  }
  cerr << "Accepted cmd connection from " << net_cmd->RemoteIp() << ":" << net_cmd->RemotePort() << endl;
  // 4. close the listener
  net_listener.Close();



  // read request
  string filename("");       // the name of the file to play
  do {
    // 1. wait for incoming data
    do {
      io_select.Clear();
      io_select.Add(*net_cmd, false /* waiting for input*/);
      io_select.Select(100); // wait 100ms
    } while (!io_select.CanRead(*net_cmd));

    // 2. :XXX: todo: error: connection closed?

    // 3. read the data from the socket
    char buffer[1024];
    size_t bytes_read(0);
    if (net_cmd->Read(buffer, 1024, bytes_read) != -1) {
      // 4. if we did not have an error ...
      // append read data to filename string
      filename.append(string(buffer, bytes_read));
    }
    // 5. do this until we read a newline character
  } while (filename[filename.length()-1] != '\n');
  // 6. chomp() filename
  while ((filename[filename.length()-1] == '\n') ||
         (filename[filename.length()-1] == '\r')) {
    filename = filename.substr(0, filename.length()-1);
  }




  // find a suitable port for listening for data connection
  while (!net_listener.Listen(data_port, 0)) {
    // 1. if we cannot listen on this port
    cout << "Cound not listen on data port " << data_port << ": " << strerror(errno) << endl;
    // try next one
    data_port++;
  }
  // 2. error handling if no free port found!
  if (!net_listener.Listen(data_port, 0)) {
    cerr << "ERROR: Problems opening server port " << data_port << " for data transmission: " << strerror(errno) << endl;
    return 1;
  }



  // answer on the request with the port for the data connection
  // 1. wait for write possibility
  do {
    io_select.Clear();
    io_select.Add(*net_cmd, true /* waiting for possibility to output */);
    io_select.Select(100);
  } while (!io_select.CanWrite(*net_cmd));
  // 2. write listening port (as charaters!)
  char data_port_cstring[7];
  size_t bytes_writen(snprintf(data_port_cstring, 7, "%d\n", data_port));
  net_cmd->Write(data_port_cstring, bytes_writen);
  // :XXX: todo: making sure that all data is written



  // accept incoming data connection
  // 1. wait for incomming connection
  do {
    io_select.Clear();
    io_select.Add(net_listener, false /* waiting for input*/);
    io_select.Select(100); // wait 100ms
  } while (!io_select.CanRead(net_listener));
  // 2. accept the connection
  auto_ptr<cTBSocket> net_data(net_listener.Accept());
  while (!net_data.get()) {
    cout << "Problems accepting data incoming connection: " << strerror(errno) << endl;
    net_data.reset(net_listener.Accept());
  }
  cerr << "Accepted data connection from " << net_data->RemoteIp() << ":" << net_data->RemotePort() << endl;
  // 4. close the listener
  net_listener.Close();



  // now we have all connection via network established:
  // 1. client has connected to cmd_port and send us the filename
  // 2. we have opened a second port for the data connection and
  // 3. send the port number to the client as answer for his request
  // 4. client has connected to the data port
  


  // open/create the pipes for mplayer
  // :XXX: this should be done at program start to avoid habing errors here
  if (pipe(inpipe) == -1) {
    cerr << "ERROR: pipe failed for inpipe: (" << errno << ") " << strerror(errno) << endl;
    return 1;
  }
  if (pipe(outpipe) == -1) {
    cerr << "ERROR: pipe failed for outpipe: (" << errno << ") " << strerror(errno) << endl;
    return 1;
  }


  // fork a child
  pid = fork();
  if (pid == -1) {
    cerr << "ERROR: fork failed: (" << errno << ") " << strerror(errno) << endl;
    return 1;
  }


  // the child: exec a mplayer
  if (pid == 0) {
    // 1. close other side's pipe ends
    close(inpipe[1]);
    close(outpipe[0]);

#ifdef DEBUG
#warning Using command output on stderr!
    cerr << "mplayer.sh" << " "
         << filename.c_str() << " "
         << "SLAVE" << endl;
#endif // DEBUG

    // 2. dup stdio to this side's pipe ends
    if((dup2(inpipe[0],  STDIN_FILENO)  < 0) ||
       (dup2(outpipe[1], STDOUT_FILENO) < 0) ||
       (dup2(outpipe[1], STDERR_FILENO) < 0)) {
      cerr << "ERROR: dup2() failed in MPlayer child: (" << errno << ") " << strerror(errno) << endl;
      exit(127);
    }
    // 3. close this side's pipes as the are dup'ed to std-io
    close(inpipe[0]);
    close(outpipe[1]);

    // 4. the final command
    execlp("mplayer.sh",
           "mplayer.sh",
           (const char*) filename.c_str(),
           "SLAVE",
           NULL);

    // 5. error handling
    cerr << "Couldn't start mplayer.sh: " << strerror(errno) << endl;
    exit(127);
  }
  // end of child code



  // the father
  //
  // if we do not sleep some time, the forked 
  // child hangs, don't know why
  usleep(10000);



  // close other side's pipe ends
  close(inpipe[0]);
  close(outpipe[1]);

  // make mplayer output pipe nonblocking
  fcntl(outpipe[0], F_SETFL, O_NONBLOCK);


  // open/create TB-objects for the communication with mplayer
  mp_input.Open(inpipe[1]);
  mp_output.Open(outpipe[0]);
  // ... and stdin
  std_input.Open(STDIN_FILENO);



  // main loop
  // 1. as long as our mplayer child is alive
  bool running(waitpid(pid, 0, WNOHANG) != pid);
  while (running) {
    running = (waitpid(pid, 0, WNOHANG) != pid);

    // 2. build up a select
    // 2.1 clear all
    io_select.Clear();
    // 2.2 add mplayer's output (INPUT for us!)
    io_select.Add(mp_output, false /* waiting for input! */);
    // 2.3 add mplayer's data fifo (INPUT for us!)
    if (fifo_input.IsOpen()) {
      // ... only if it is open yet!
      io_select.Add(fifo_input, false);
    }
    // 2.4 add std input
    io_select.Add(std_input, false /* waiting for input! */);
    // 2.5 add net cmd input
    io_select.Add(*net_cmd, false);
    // 3. select the descriptors
    if (io_select.Select(100) > 0) {

      // at least one of the has some data
      // 1. most important is mplayer's output
      //    as if it is blocked, this can lead to a deadlock!
      if (io_select.CanRead(mp_output)) {
        // 1.1 read what mplayer wants us to tell!
        char mplayer_output[2048];
        size_t bytes_read(0);
        if (mp_output.Read(mplayer_output, 2048, bytes_read) != -1) {
          // 1.2 if we did not have an error ...

          // ... write what we've read
          // ... cerr (debugging only)
          cout.write(mplayer_output, bytes_read);
          cout.flush();
          // ... to the command connection
          while (bytes_read > 0) {
            net_cmd->Write(mplayer_output, bytes_read);
          }

          // if we have read something and the mplayer fifo for
          // data output is not open yet, we open it now
          // if we try this earlier, we might get blocked while opening
          if (!fifo_input.IsOpen()) {
            fifo_input.Open("grab.mpg", O_RDONLY + O_NONBLOCK);
          }
        }
      }

      string next_command("");

      // 2. next important: commands from the network connection
      if (io_select.CanRead(*net_cmd)) {
        // 2.1 read the command
        char buf[256];
        size_t read_bytes(0);
        net_cmd->Read(buf, 256, read_bytes);

        // 2.2 put it in the next-command string
        next_command = string(buf, read_bytes);
        // 2.3 find newline in the string
        if (next_command.find('\n') == string::npos) {
          // ... if not: ignore command
          // :XXX: to buffered reading over multiple tries
          next_command = string("");
        } else {
          // chomp() the string
          while ((next_command[next_command.length()-1] == '\n') ||
                 (next_command[next_command.length()-1] == '\r')) {
            next_command = next_command.substr(0, next_command.length()-1);
          }
        }
      }

      // 3. for debugging: commands from stdin
      if (io_select.CanRead(std_input)) {
        // 3.1 read the command
        char buf[256];
        size_t read_bytes(0);
        std_input.Read(buf, 256, read_bytes);

        // 3.2 put it in the next-command string
        next_command = string(buf, read_bytes);
        // 3.3 find newline in the string
        if (next_command.find('\n') == string::npos) {
          // ... if not: ignore command
          // :XXX: to buffered reading over multiple tries
          next_command = string("");
        } else {
          // chomp() the string
          while ((next_command[next_command.length()-1] == '\n') ||
                 (next_command[next_command.length()-1] == '\r')) {
            next_command = next_command.substr(0, next_command.length()-1);
          }
        }
      }

      // 4. if we have got a next command
      if (!next_command.empty()) {
        // write it to mplayer's input
        size_t length(next_command.length()+1);
        mp_input.Write((char*) (next_command + string("\n")).c_str(), length);
      }

      // 5. if the fifo has some data for us (only test it if it is already open!)
      if (fifo_input.IsOpen() && io_select.CanRead(fifo_input)) {
        // 5.1 read the data
        char buffer[4096];
        size_t bytes_read(0);
        fifo_input.Read(buffer, 4096, bytes_read);
        // 5.2 and write it to the data network connection
        while (bytes_read > 0) {
          // all of it!
          do {
            io_select.Clear();
            io_select.Add(*net_data, true /* waiting for possibility to output */);
            io_select.Select(100);
          } while (!io_select.CanWrite(*net_data));
          net_data->Write(buffer, bytes_read);
        }
      }

    } // end if: data available on one of the connections
  } // end while: main loop



  // shutdown
  // 1. wait until mplayer child died (otherwise we get a zoobie)
  if (waitpid(pid, 0, WNOHANG) != -1) {
    // child not terminated yet
    cerr << "If it won't die";
    // as long as it did not terminate
    while (waitpid(pid, 0, WNOHANG) == 0) {
      usleep(100000); // wait 100ms
      cerr << ", we'll shoot it";
      // and KILL it!
      kill(pid, SIGTERM);
    }
//     cerr << endl;
  }








  unlink("grab.mpg");
  rmdir(tmp_dir);

  cerr << "exit" << endl;

  return 0;
}
